
(* Generate an interpolant from a refutation of literal sets (a,b).
   
   Note this only implements the theory rules from the paper, not
   the propositional reasoning rules.

   Also note this implementation is inefficiant. Various things need
   to be memoized to make it efficient. *)
      
module Set = PSet.Set

open Num
open Term
open Predicate

(* Proofs *)

type proof =
    Hyp of predicate
  | Comb of num * proof * num * proof
  | Refl of term
  | Symm of proof
  | Trans of proof * proof
  | Cong of string * (proof list)
  | EqNeq of proof
  | LeqEq of proof
  | EqLeq of term * term * proof * proof

(* Interpolation *)

let interpolate (a,b) refutation =
    
   (* is literal p contained in set a? *)

  let pred_in_A p = Set.mem p a in
    
  (* is every symbol in term f also in B? (i.e., f \preceq B) *)
    
  let rec term_in_B f = term_in_pred_set f b in

  (* interpolation rules that generate inequalities *)

  let rec ineq_rules =
    function
      | Hyp((Inequality(x)) as p) ->
	  if pred_in_A p
	  then (x,(x,pTrue,pTrue)) (* HypLeq-A *)
	  else (x,(zero,pTrue,pTrue)) (* HypLeq-B *)
      | Hyp(Not(Inequality(x)) as p) ->  
	  let x = sub_term (term_of_int(-1)) x in  (* translate ~(0<=x) to 0<=-x-1 *) 
	    if pred_in_A p
	    then (x,(x,pTrue,pTrue)) (* HypLeq-A *)
	    else (x,(zero,pTrue,pTrue)) (* HypLeq-B *)
      | Comb(c1,pf1,c2,pf2) ->
	  let (x,(x',rho,gamma)) = ineq_rules pf1 in
	  let (y,(y',rho',gamma')) = ineq_rules pf2 in
	    ((add_term (mult_term c1 x) (mult_term c2 y)),
	     ((add_term (mult_term c1 x') (mult_term c2 y')),
	      pAnd(rho,rho'),
	      pAnd(gamma,gamma'))) (* Comb *)
      | LeqEq(pf) ->
	  let ((y,x),(y',x',rho,gamma)) = eq_rules pf in
	    if not(y' = x)
	    then ((sub_term x y),((sub_term (add_term x y') (add_term x' y)),rho,gamma)) (* LeqEq *)
	    else ((sub_term x y),((sub_term x y),rho,gamma)) (* LeqEq' *)
      | EqNeq(pf) ->
	  let ((x,y),(x',y',rho,gamma)) = eq_rules pf in
	    if pred_in_A (pNot(pEquality(x,y))) then
	      if not(y' = x && x' = y) then
		((term_of_int (-1)),(zero,rho,pAnd(gamma,pNot(pEquality(x',y'))))) (* EqNeq-A *)
	      else ((term_of_int (-1)),(zero,rho,pFalse))  (* EqNeq-A' *)
	    else ((term_of_int (-1)),(zero,rho,gamma))
      | _ -> failwith "interpolate: unknown inequality rule"
	  
  (* interpolation rules that generate equalities *)

  and eq_rules = 
    function
      | Hyp((Equality(x,y)) as p) ->
	  if pred_in_A p
	  then let xB = term_in_B x and yB = term_in_B y in
	    ((x,y),((if xB then x else y),
		    (if yB then y else x),
		    pTrue,
		    (if xB && yB then pEquality(x,y) else pTrue))) (* HypEq *)
	  else ((x,y),(x,y,pTrue,pTrue)) (* HypEq-B *)
      | Refl(x) ->
	  ((x,x),(x,x,pTrue,pTrue)) (* Refl *)
      | Symm(pf) ->
	  let ((x,y),(x',y',rho,gamma)) = eq_rules pf in
	    ((y,x),(y',x',rho,gamma)) (* Symm *)
      | Trans(pf1,pf2) ->
	  let ((x,y),(x',y',rho,gamma)) = eq_rules pf1 in
	  let ((_,z),(y'',z',rho',gamma')) = eq_rules pf2 in
	    if not(x' = y) && not(z' = y) then
	      ((x,z),(x',z',pAnd(rho,rho'),pAnd(pAnd(gamma,gamma'),(eq y' y'')))) (* Trans *)
	    else
	      ((x,z),((if x' = y then y'' else x'),(if z' = y then y' else z'),
		      pAnd(rho,rho'),pAnd(gamma,gamma'))) (* Trans' *)
      | Cong(f,pfs) ->
	  let itps = List.map eq_rules pfs in
	  let fxbar = apply (f,List.map (function ((x,_),_) -> x) itps) in
	  let fybar = apply (f,List.map (function ((_,y),_) -> y) itps) in
	  let fxbar' = apply (f,List.map (function (_,(x,_,_,_)) -> x) itps) in
	  let fybar' = apply (f,List.map (function (_,(_,y,_,_)) -> y) itps) in
	  let conj_rho = conj_list (List.map (function (_,(_,_,rho,_)) -> rho) itps) in
	  let conj_gamma = conj_list (List.map (function (_,(_,_,_,gamma)) -> gamma) itps) in
	    if (term_in_B fxbar) || (term_in_B fybar) then ((fxbar,fybar),(fxbar',fybar',conj_rho,conj_gamma))
	    else
	      let foo = conj_list (List.map (function (_,(x',y',_,gamma)) ->
					       if (term_in_B x') && (term_in_B y')
					       then implies gamma (pEquality(x',y')) else pTrue) itps) in
		((fxbar,fybar),(fybar,fxbar,pAnd(conj_rho,foo),conj_gamma))
      | EqLeq(x,y,pf1,pf2) -> (* pf1 proves y <= x, and pf2 proves x <= y *)
	  let (_,(x',rho,gamma)) = ineq_rules pf1 in
	  let (_,(y',rho',gamma')) = ineq_rules pf2 in
	    if (term_in_B x) && (term_in_B y)
	    then ((x,y),(x,y,pAnd(rho,rho'),pAnd(pAnd(gamma,gamma'),pAnd(pLeq(zero,x'),pLeq(zero,y'))))) (* EqLeq-BB *)
	    else if not(term_in_B x) && (term_in_B y)
	    then ((x,y),(add_term x y',y,
			 pAnd(pAnd(rho,rho'),pLeq(zero,minus_term(add_term x' y'))),
			 pAnd(pAnd(gamma,gamma'),pLeq(zero,add_term x' y')))) (* EqLeq-AB *)
	    else if (term_in_B x) && not (term_in_B y)
	    then ((x,y),(x,add_term y x',
			 pAnd(pAnd(rho,rho'),pLeq(zero,minus_term(add_term x' y'))),
			 pAnd(pAnd(gamma,gamma'),pLeq(zero,add_term x' y')))) (* EqLeq-BA *)
	    else ((x,y),(y,x,
			 pAnd(pAnd(rho,rho'),pAnd(pLeq(zero,sub_term y (add_term x y')),pLeq(zero,sub_term x (add_term y x')))),
			 pAnd(gamma,gamma')))                                   (* EqLeq-AA *)
      | _ -> failwith "interpolate: unknown equality rule"
  in
    match ineq_rules refutation with (* "false" is represented by the inequality 0 <= -1 *)
	(_,(x',rho,gamma)) -> pOr(pNot(rho),pAnd (pLeq(zero,x'), gamma))

