(* Terms

   This module provides operations on linear arithmetic terms,
   maintaining them in normal form. A arithmetic term ("term")
   is a linear combination of functional terms ("funterm").

 *)

(* use polymorhic Set and Map packages *)

module Set = PSet.Set
module Map = PMap.Map

open Map
open Num (* this modules represents rational numbers *)
	       
(* our data type -- funterm is either a variable or
   a function applied to terms, while a term is numeric
   constant plus a map from functional terms to their
   (non-zero) coefficients *)

type funterm =       
    Variable of string
  | Application of string * (term list)
      
and term =
    num * ((funterm,num) t)  (* note t is Map.t *)
      

(* multiply a term by an integer coefficient *)

let mult_term coeff (const, comb) =
  ((mult_num coeff const),
   map (mult_num coeff) comb)
  
(* add two terms, reducing result to normal form *)

let add_term (const1, comb1) (const2, comb2) =

  let comb_filter_zeros comb =
    PMap.Map.fold ~f:(fun ~key:t ~data:c m ->
			if not(c = (num_of_int 0)) then
			  add t c m
			else m) comb ~init:(empty) in
    
  let comb_add =
    fold ~f:(fun ~key:t ~data:c m ->
	       add t (try (add_num c (find t m)) with Not_found -> c) m) in
    
    ((add_num const1 const2),
     (comb_filter_zeros (comb_add comb1 comb2)))
    
(* arithmetic negation of a term *)

let minus_term x = mult_term (num_of_int (-1)) x
		   
(* difference of terms *)

let sub_term t1 t2 = add_term t1 (minus_term t2)
		       
(* cast a functional term to an arithmetic term *)

let term_of_funterm ft =
  ((num_of_int 0),(add ft (num_of_int 1) (empty)))
  
(* cast an numeric constant to a term *)

let term_of_num : (num -> term) =
  fun n -> (n,empty)
    
(* cast an int to a term *)

let term_of_int i = term_of_num (num_of_int i)

(* the constant zero term *)

let zero = term_of_num (num_of_int 0)

(* apply a function to two terms *)		  

let apply (f, args) =
  term_of_funterm (Application (f,args))

(* cast a string to a variable *)

let var s = term_of_funterm (Variable s)    

(* get the set of symbols occuring in a term *)

let rec support_term (_,t) =
  fold ~f:(fun ~key:t ~data:_ s -> Set.union (support_funterm t) s) ~init:Set.empty t
and support_funterm =
  function
      Variable n -> Set.singleton n
    | Application (f,args) ->
	List.fold_left Set.union (Set.singleton f) 
	(List.map support_term args)

(* printer functions for terms *)

let rec print_funterm fmt =
  function
      Variable(s) -> Format.fprintf fmt "%s" s
    | Application(s,l) -> 
	begin
	  Format.fprintf fmt "%s( " s;
	  List.iter (print_term fmt) l;
	  Format.fprintf fmt ")";
	end

and  print_term fmt (const, comb) =
  let first = ref true in
    begin
      if not(const = (num_of_int 0))
      then 
	(Format.fprintf fmt "%s" (string_of_num const); first := false);
      iter ~f:(fun ~key:x ~data:y ->
	      begin
		if not !first then Format.fprintf fmt " + ";
		first := false;
		if not(y = (num_of_int 1))
		then Format.fprintf fmt "%s * " (string_of_num y);
		print_funterm fmt x
	      end) comb;
      if !first then Format.fprintf fmt "0"
    end
